/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** Stellt eine Spielfigur dar.
 * \author Martin Gernhard
 */

#include <iostream>
#include <string>
#include <sstream>
#include <fstream>

#include <SDL.h>
#include <SDL_opengl.h>

#include "logger.h"
#include "sprite.h"
#include "figur.h"
#include "utils.h"
#include "tile.h"
#include "level.h"
#include "herzkiste.h"

#include "ring.h"
#include "stachel.h"
#include "punkte.h"

Figur::Figur( const std::string name, const std::string& animationsdatei, Level* level,
	SpriteManager* spriteManager )
{
	this->name = name;
	this->dateiname = animationsdatei;
	this->animator = new Animator(animationsdatei, spriteManager);
	this->level = level;
	this->spriteManager = spriteManager;

	this->abstandLinks  = 0;
	this->abstandRechts = 0;
	this->abstandOben   = 0;
	this->abstandUnten  = 0;

	this->dx = 0;
	this->dy = 0;

	this->gespiegelt = false;
	this->faellt_ = false;
	this->tot = false;

	this->schaden = 1;
	this->verschiebung = 0;
	this->unverwundbarkeitsZaehler = 0;

	this->punkte = Punkte::PUNKTE_GEGNER_NORMAL;
	this->springKillbar = true;
	this->gravitation = GRAVITATION;
	this->typ = FigurTyp::FIGURTYP_STANDARD;
}

Figur::~Figur()
{
	delete animator;
	
	for (std::list<Steuerfunktion*>::iterator iter = steuerfunktionen.begin(); iter != steuerfunktionen.end();){
		Steuerfunktion* stf = *iter;
		logger->debug("Steuerfunktion '" + stf->getName() + "' von Figur '"
					+ name + "' entfernt.");
		iter = steuerfunktionen.erase(iter);
		delete stf;			
	}
	
	logger->debug("Figur '" + name + "' entfernt.");
}

void Figur::zeichnen()
{
	if (unverwundbarkeitsZaehler %2 == 1){
		return;
	}

	const int maxVerschiebung = level->getBreite() - SCREEN_WIDTH;
	
	if (verschiebung < 0 ){
		animator->zeichnen(x, y);
	}else if ( verschiebung > maxVerschiebung ){
		animator->zeichnen(x - maxVerschiebung, y);
	}else{
		animator->zeichnen(x - verschiebung, y);
	}
}

void Figur::kollision(Richtung richtung)
{
	Kollision kollision(richtung);
	this->kollision(kollision);
}

void Figur::kollision(Kollision kollision)
{
	std::list<Steuerfunktion*>::iterator iter;
	
	for (iter = steuerfunktionen.begin(); iter != steuerfunktionen.end();){
		Steuerfunktion* stf = *iter;
		stf->kollision(kollision);

		if (stf->istBeendet()){
			iter = steuerfunktionen.erase(iter);
			delete stf;			
		}else{
			iter++;
		}
	}
}

void Figur::naechsterZustand()
{
	beruehrteKisten.clear();
	std::list<Steuerfunktion*>::iterator iter;
	
	for (iter = steuerfunktionen.begin(); iter != steuerfunktionen.end();){
		Steuerfunktion* stf = *iter;

		stf->ausfuehren();

		if (stf->istBeendet()){
			iter = steuerfunktionen.erase(iter);
			delete stf;			
		}else{
			iter++;
		}
	}

	// Tile-Kollisionen berechnen, auflsen und an die Steuerfunktionen
	// weiterleiten.
	int tileKoordinateX;	
	if (dx > 0){
		if ( kollisionVertikal(x +dx + (getBreite()-1) - getAbstandRechts(), y + getAbstandOben(), tileKoordinateX) ){
			x = tileKoordinateX * Tile::TILE_BREITE - (getBreite()) - 1 + getAbstandRechts();
			this->kollision(Richtung(RICHTUNG_RECHTS));
			animator->setzeBewegungsmodus(STEHEN);
		}else{
			x += dx;
			animator->setzeBewegungsmodus(rennt ? RENNEN : GEHEN);
		}
		animator->setzeRichtung(RICHTUNG_RECHTS);
	}else if (dx < 0){
		if ( kollisionVertikal(x + dx + getAbstandLinks(), y + getAbstandOben(), tileKoordinateX) ){
			x = (tileKoordinateX+1) * Tile::TILE_BREITE + 1 - getAbstandLinks();
			this->kollision(Richtung(RICHTUNG_LINKS));
			animator->setzeBewegungsmodus(STEHEN);
		}else{
			x += dx;
			animator->setzeBewegungsmodus(rennt ? RENNEN : GEHEN);
		}
		animator->setzeRichtung(RICHTUNG_LINKS);
	}else{
		animator->setzeBewegungsmodus(STEHEN);
	}

	int tileKoordinateY;
	if (dy < 0){
		// Figur springt.
		if (kollisionHorizontal(x + getAbstandLinks(), y + dy + getAbstandOben(), tileKoordinateY) ){
			y = (tileKoordinateY+1) * Tile::TILE_HOEHE + 1 - getAbstandOben();
			this->kollision(Richtung(RICHTUNG_OBEN));
			faellt_ = false;
			animator->setzeBewegungsmodus(SPRUNG_AUF);
		}else{
			y += dy;
			dy += gravitation;
			dy = max ((0 - Tile::TILE_HOEHE) + 1, dy);
			faellt_ = false;
			animator->setzeBewegungsmodus(SPRUNG_AUF);
		}
	}else{
		if (kollisionHorizontal(x + getAbstandLinks(), y + dy + (getHoehe()) - getAbstandUnten(), tileKoordinateY) ){
			// Figur befindet sich auf dem Boden.
			y = tileKoordinateY * Tile::TILE_HOEHE - getHoehe() + getAbstandUnten() ;
			this->kollision(Richtung(RICHTUNG_UNTEN));
			faellt_ = false;
//				dy = 0;
		}else{
			// Figur fllt.
			y += dy;
			dy += gravitation;
			dy = min (Tile::TILE_HOEHE - 1, dy);
			faellt_ = true;
			if (dy < 0){
				animator->setzeBewegungsmodus(SPRUNG_AB);
			}
		}
	}

	/* Figur innerhalb der Grenzen halten */
	
	// Linker Rand
	if ( x <  0 ){
		this->kollision(Richtung(RICHTUNG_LINKS));
	}
	x = max(0, x);
	
	// Rechter Rand
	// FIXME: HACK HACK HACK ausgeschaltet fr "Levelende"
	/*if ( x >  level->getBreite() - getBreite()){
		this->kollision(Richtung(RICHTUNG_RECHTS));
	}
	x = min(level->getBreite() - getBreite(), x);*/
	
	// Unterer Rand
	if ( y >= level->getHoehe() ){
		// Figur ist unten aus dem Bild gefallen.
		tot = true;
	}
	// Hier keine Begrenzung durchfhren, soll ruhig rausfallen.
	
	// Unverwundbarkeitszhler wird mit jedem Frame erniedrigt.
	unverwundbarkeitsZaehler--;
	
	animator->naechstesBild();
}

int Figur::getBreite()
{
//	return animator->aktuellesBild()->getBreite();
	return animator->erstesBild()->getBreite();
}
	
int Figur::getHoehe()
{
//	return animator->aktuellesBild()->getHoehe();
	return animator->erstesBild()->getHoehe();
}

void Figur::setzePosition(int x, int y)
{
	this->x = x;
	this->y = y;
}

void Figur::setX(int x)
{
	this->x = x;
}

void Figur::setY(int y)
{
	this->y = y;
}

int Figur::getX()
{
	return x;
}

int Figur::getY()
{
	return y;
}

Kollision Figur::pruefeKollision(const Figur& figur2){

	/* Beginn "Bounding Box" - Test */
	const int links  = x;
	const int rechts = links + animator->aktuellesBild()->getBreite();
	const int oben   = y;
	const int unten  = oben + animator->aktuellesBild()->getHoehe();

	const int links2  = figur2.x;
	const int rechts2 = links2 + figur2.animator->aktuellesBild()->getBreite();
	const int oben2   = figur2.y;
	const int unten2  = oben2 + figur2.animator->aktuellesBild()->getHoehe();

	// 2 ber 1?
	if (unten2 < oben){ return false; }
	// 2 links von 1?
	if (rechts2 <  links){ return false; }
	// 2 unter 1?
	if (unten < oben2){ return false; }
	// 2 rechts von 1?
	if (rechts < links2){ return false; }

	/*
	 * "Bounding Box" - Test war positiv, genauer prfen.
	 * Vorgehen: zuerst das berschneidungsrechteck (relativ zum Bildschirm)
	 * berechnen.
	 */

	// Die rechteste linke Kante ist die linke Kante der berlappung.
	const int u_links = max(links, links2);
	// Die linkeste rechte Kante ist die rechte Kante der berlappung.
	const int u_rechts = min(rechts, rechts2);
	// Die unterste Oberkante ist die Oberkante der berlappung.
	const int u_oben = max(oben, oben2);
	// Die oberste Unterkante ist die Unterkante der berlappung.
	const int u_unten = min(unten, unten2);

	/* Die berschneidungsrechtecke vergleichen. */
	SDL_Surface* bild1 = animator->aktuellesBild()->getBilddaten();
	SDL_Surface* bild2 = figur2.animator->aktuellesBild()->getBilddaten();
	
	SDL_LockSurface(bild1);
	SDL_LockSurface(bild2);

	// Die beiden Rechtecke innerhalb der Sprites, d.h. relativ zu ihrer
	// oberen, linken Ecke, berechnen.
	
	const int b1_links	= u_links	-	x;
	const int b1_oben	= u_oben	-	y;

	const int b2_links	= u_links	-	figur2.x;
	const int b2_oben	= u_oben	-	figur2.y;

	for (int y = 0; y < ((u_unten - u_oben)+1); y++){
		for (int x = 0; x < ((u_rechts - u_links)+1); x++){
			int position1 = (x + b1_links) + (y + b1_oben) * bild1->w;
			int position2 = (x + b2_links) + (y + b2_oben) * bild2->w;

			Uint32 pixel1 = *((Uint32*)bild1->pixels + position1);
			Uint32 pixel2 = *((Uint32*)bild2->pixels + position2);

			Uint32 alpha1 = pixel1 & bild1->format->Amask;
			Uint32 alpha2 = pixel2 & bild2->format->Amask;

			if (alpha1 && alpha2){
				SDL_UnlockSurface(bild1);
				SDL_UnlockSurface(bild2);
				
				Kollision kollision(true);
				kollision.setKollidiertMitGegner(true);

				// Hier bestimmen, ob das ein Sprung auf den Kopf war
				if ( (((oben2 + unten2) / 2) > u_unten) && (faellt_)){
					kollision.richtungHinzufuegen(RICHTUNG_UNTEN);
				}

				// Hier bestimmen, ob das ein Sprung unter einen zerstrbaren
				// Block war.
				if ( (((oben2 + unten2) / 2) < u_oben) && (!faellt_)){
					kollision.richtungHinzufuegen(RICHTUNG_OBEN);
				}
		
				return kollision;
	
			}
		}
	}

	SDL_UnlockSurface(bild1);
	SDL_UnlockSurface(bild2);

	return false;
}

void Figur::steuerfunktionHinzufuegen(Steuerfunktion* steuerfunktion){
	steuerfunktionen.push_back(steuerfunktion);
}

void Figur::setzeAbstaende(int abstandLinks, int abstandOben, int abstandRechts,
			int abstandUnten)
{
	this->abstandLinks  = abstandLinks;
	this->abstandOben   = abstandOben;
	this->abstandRechts = abstandRechts;
	this->abstandUnten  = abstandUnten;
}

void Figur::setzeBoundingBox(int bxl, int byt, int bxr, int byb)
{
	this->abstandLinks  = bxl;
	this->abstandOben   = byt;
	this->abstandUnten  = getHoehe() - ( byb + 1);
	this->abstandRechts = getBreite() - ( bxr + 1);
	
	logger->debug(itos(abstandLinks) + ":" + itos(abstandRechts) + ":"
			+ itos(abstandOben) + ":" + itos(abstandUnten));
}

bool Figur::kollisionVertikal( int x, int y, int &tileKoordinateX )
{
	// Linke Seite des Tiles in Pixeln
	int tilePositionY = y - (y % Tile::TILE_HOEHE);
	int testEndeY = y + (getHoehe() - 1) - (getAbstandUnten() + getAbstandOben());
	
	tileKoordinateX = x / Tile::TILE_HOEHE;
	int tileKoordinateY = y / Tile::TILE_HOEHE;

	
	while (tilePositionY <= testEndeY){
		if ( level->isSolid(tileKoordinateX, tileKoordinateY)
			|| level->sucheKisteAnPosition(tileKoordinateX, tileKoordinateY) ){
			return true;
		}
		
		tileKoordinateY++;
		tilePositionY += Tile::TILE_HOEHE;
	} 
	
	return false;
}

bool Figur::kollisionHorizontal( int x, int y, int &tileKoordinateY )
{
	// Linke Seite des Tiles in Pixeln
	int tilePositionX = x - (x % Tile::TILE_BREITE);
	int testEndeX = x + (getBreite() - 1) - (getAbstandRechts() + getAbstandLinks());
	
	int tileKoordinateX = x / Tile::TILE_HOEHE;
	tileKoordinateY = y / Tile::TILE_BREITE;


	bool wandBeruehrt = false;
	bool kisteBeruehrt = false;
	
	while (tilePositionX <= testEndeX){
		if ( level->isSolid(tileKoordinateX, tileKoordinateY) ){
			wandBeruehrt = true;
		}
		
		Kiste* kiste = level->sucheKisteAnPosition(tileKoordinateX, tileKoordinateY);
		if (kiste != NULL){
			beruehrteKisten.push_back( kiste );
			kisteBeruehrt = true;
		}

		tileKoordinateX++;
		tilePositionX += Tile::TILE_HOEHE;
	} 

	return (kisteBeruehrt || wandBeruehrt);
}

int Figur::getDX()
{
	return dx;
}

int Figur::getDY()
{
	return dy;
}

void Figur::setDX(int dx)
{
	this->dx = dx;

	if (dx < 0){
		gespiegelt = true;
	}else if (dx > 0){
		gespiegelt = false;
	}
}

void Figur::setDY(int dy)
{
	this->dy = dy;
}

void Figur::setVerschiebung(int verschiebung)
{
	this->verschiebung = verschiebung;
}

bool Figur::faellt()
{
	return faellt_;
}

bool Figur::isTot()
{
	return tot;
}

void Figur::setTot(bool tot)
{
	this->tot = tot;
}

std::string Figur::getName()
{
	return name;
}

FigurTyp Figur::getTyp()
{
	return typ;
}

bool Figur::isUntenRausgefallen()
{
	return y > level->getHoehe() + HUD_HOEHE;	
}

void Figur::setRennt(bool rennt)
{
	this->rennt = rennt;
}

void Figur::setzeUnverwundbar(int frames)
{
	this->unverwundbarkeitsZaehler = frames;
}

bool Figur::isUnverwundbar()
{
	return (unverwundbarkeitsZaehler > 0);
}

int Figur::getAbstandLinks()
{
	if (gespiegelt){
		return abstandRechts;
	}else{
		return abstandLinks;
	}
}

int Figur::getAbstandRechts()
{
	if (gespiegelt){
		return abstandLinks;
	}else{
		return abstandRechts;
	}
}

int Figur::getAbstandOben()
{
	return abstandOben;
}

int Figur::getAbstandUnten()
{
	return abstandUnten;
}

std::string Figur::alsString()
{
	std::stringstream output;
	
	output << "FIG: " << typ.alsString() << std::endl;
	output << "X: " << x << std::endl;
	output << "Y: " << y << std::endl;
	output << "ANI: " << dateiname << std::endl;
	
	for (std::list<Steuerfunktion*>::iterator iter = steuerfunktionen.begin(); iter != steuerfunktionen.end(); iter++){
		Steuerfunktion* stf = *iter;
		output << "STF: " << stf->alsString() << std::endl;
	}

	for (std::list<std::string>::iterator iter = parameter.begin(); iter != parameter.end(); iter++){
		std::string param = *iter;
		output << "PAR: " << param << std::endl;
	}


	return output.str();
}

Figur* Figur::ausString(SpriteManager* spriteManager, Level* level,
			std::string input_)
{
	logger->debug("<FigurAusString>");
	// FIXME: Einzelzeilen-Stringstreams sind unntig, getline akzeptiert ein Trennzeichen);
	// Gleich viele einzelne Gets?
	std::istringstream input(input_);
	std::string beschreibung;
	std::string zeile_;

	// Figurtyp
	getline(input, zeile_);

	std::vector<std::string> typZeile = parse(zeile_, 2);
	std::string typ = typZeile[1];
	
	logger->debug("Typ = '" + typ + "'");

	// X-Position
	getline(input, zeile_);
	std::istringstream zeile2(zeile_);

	int x = 0;
	zeile2 >> beschreibung;
	zeile2 >> x;

	logger->debug("X = '" + itos(x) + "'");

	// Y-Position
	getline(input, zeile_);
	std::istringstream zeile3(zeile_);

	int y = 0;
	zeile3 >> beschreibung;
	zeile3 >> y;

	logger->debug("Y = '" + itos(y) + "'");

	// Animationsdatei
	getline(input, zeile_);
	std::vector<std::string> animationsdateiZeile = parse(zeile_, 2);
	std::string animationsdatei = animationsdateiZeile[1];
	
	logger->debug("Animationsdatei = '" + animationsdatei + "'");

	//////////////////////////
	std::string name(animationsdatei);
	//////////////////////////

	Figur* figur;
	
	if (typ == FigurTyp::FIGURSTRING_STANDARD){
		figur = new Figur(name, animationsdatei, level, spriteManager);
	}else if (typ == FigurTyp::FIGURSTRING_RING){
		figur = new Ring(name, animationsdatei, level, spriteManager);
	}else if (typ == FigurTyp::FIGURSTRING_STACHEL){
		figur = new Stachel(name, animationsdatei, level, spriteManager);
	}else if (typ == FigurTyp::FIGURSTRING_KISTE){
		Kiste* kiste = new Kiste(name, animationsdatei, level, spriteManager);
		level->kisteHinzufuegen(kiste);
		figur = kiste;
	}else if (typ == FigurTyp::FIGURSTRING_HERZKISTE){
		Kiste* kiste = new HerzKiste(name, animationsdatei, level, spriteManager);
		level->kisteHinzufuegen(kiste);
		figur = kiste;
	}else{
		throw "Unbekannter Figurtyp";
	}
	
	figur->setzePosition(x, y);

	while(!input.eof()){
		getline(input, zeile_);

		if (zeile_ != ""){
			// Teilt in "STF:", "PAR:" etc. und den Rest der Zeile
			std::vector<std::string> parameter = parse(zeile_, 1);

			if (parameter[0] == "STF:"){
				figur->steuerfunktionHinzufuegen(Steuerfunktion::ausString(figur, parameter[1]));
			}else if (parameter[0] == "PAR:"){
				figur->setParameter(parameter[1]);
			}
		}
	}

	logger->debug("</FigurAusString>");

	return figur;
}

void Figur::setParameter(std::string parameter)
{
	trim( parameter );
	(this->parameter).push_back( parameter );
	logger->debug("Parameter gefunden: '" + parameter + "'");

	std::istringstream input(parameter);
	std::string name;

	input >> name;

	if (name == "Schaden"){
		input >> this->schaden;
	}

	if (name == "Gravitation"){
		input >> this->gravitation;
	}
}

// Ldt alle Gegner fr einen Level aus einer Datei.
std::list<Figur*> Figur::ladeListeAusDatei(std::string dateiname,
			SpriteManager* spriteManager, Level* level)
{
	std::list<Figur*> gegner;
	
	std::ifstream input(dateiname.c_str());
	
	if (input == NULL){
		sdlFehlerAusgeben("Objektdatei '" + dateiname + "' konnte nicht geladen werden.");
		exit(1);
	}

	std::ostringstream output;

	// Einlesen, bis eine Leerzeile gefunden wird, dann alles biher gefundene
	// an Figur::ausString bergeben und von vorne anfangen, zu lesen.
	while(!input.eof()){
		std::string zeile;
		getline(input, zeile);
		if (zeile == ""){
			if (output.str().size() != 0){
				logger->debug("<Figur-Input>");
				logger->debug(output.str());
				logger->debug("</Figur-Input>");
				Figur* figur = Figur::ausString(spriteManager, level, output.str());
				gegner.push_back(figur);

				output.str("");
				output.clear();
			}
		}else{
			output << zeile << std::endl;
		}
	}

	input.close();
	
	return gegner;
}

void Figur::speichereListeAlsDatei(std::list<Figur*> objekte, std::string dateiname)
{
	std::ofstream datei(dateiname.c_str());

	if (datei == NULL){
		logger->error("Objektdatei '" + dateiname + "' konnte nicht geschrieben werden.");
		exit(1);
	}

	for (std::list<Figur*>::iterator iter = objekte.begin(); iter != objekte.end(); iter++){
		Figur* g = *iter;
		datei << g->alsString() << std::endl;
	}

	datei.close();
}

unsigned int Figur::getSchaden()
{
	return schaden;
}

std::list<Figur*>& Figur::getBeruehrteKisten()
{
	return beruehrteKisten;
}

int Figur::getPunkte()
{
	return this->punkte;
}
