/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** Stellt eine Spielfigur dar.
 * \author Martin Gernhard
 */

#ifndef FIGUR_H
#define FIGUR_H

#include <list>
#include <vector>

#include "animator.h"
#include "steuerfunktion.h"
#include "sprite.h"
#include "kollision.h"
#include "bewegung.h"
#include "spritemanager.h"
#include "figurtyp.h"

// Vorwrtsdeklarationen
class Steuerfunktion;
class Tile;
class Level;

class Figur
{
private:

	// Die X-Position der Figur.
	int x;
	// Die Y-Position der Figur.
	int y;
	
	// Die X-Geschwindigkeit der Figur.
	int dx;
	// Die Y-Geschwindigkeit der Figur.
	int dy;

	// Die Bounding Box der Figur
	int abstandLinks;
	int abstandRechts;
	int abstandOben;
	int abstandUnten;

	// Befindet sich die Figur gerade im freien Fall?
	bool faellt_;

	// Ist die Figur verstorben?
	bool tot;

	// Rennt die Figur?
	bool rennt;
	
	// Anzahl der Frames, die die Figur noch unverwundbar sein wird.
	int unverwundbarkeitsZaehler;
	
	// Die Scroll-Verschiebung der Figur.	
	int verschiebung;
	
	// Guckt die Figur nach links und muss gespiegelt werden?
	bool gespiegelt;

	// Ein simpler Name, hauptschlich fr DEBUG-Zwecke benutzt.
	std::string name;

	// Der Level, in dem sich die Figur aufhlt.
	Level* level;

	// Steuerfunktionen, die diese Figur beeinflussen.
	std::list<Steuerfunktion*> steuerfunktionen;

	// In diesem Frame berhrte Kisten.
	std::list<Figur*> beruehrteKisten;

	// Die Anzahl der Punkte die diese Spielfigur wert ist.
	int punkte;

	bool kollisionVertikal( int x, int y, int &tileKoordinateX );
	bool kollisionHorizontal( int x, int y, int &tileKoordinateX );

	int getAbstandLinks();
	int getAbstandRechts();
	int getAbstandOben();
	int getAbstandUnten();

	// Zustzliche Parameter in Stringform, die aus der Objectdatei gelesen
	// und auch dort hinein geschrieben werden.
	std::list<std::string> parameter;

protected:
	// Zustndig fr die Verwaltung der Animationsphasen und -modi.
	Animator* animator;

	// Der zustndige SpriteManager, wird an den Animator weitergeleitet. 
	SpriteManager* spriteManager;

	// Kann die Figur durch einen Sprung auf den Kopf gettet werden?
	bool springKillbar;

	// Die Strke der Gravitation, der diese Figur unterliegt. 
	int gravitation;

	// Der Typ dieser Figur.
	FigurTyp typ;

	unsigned int schaden;

	// Der Dateiname muss mit abgespeichert werden knnen.
	std::string dateiname;

public:

	/**
	 * Erstellt eine neue Figur.
	 * \param animationsdatei Datei, welche die Animationsbeschreibungen enthlt.
	 */
	Figur( const std::string name, const std::string& animationsdatei, Level* level,
		SpriteManager* spriteManager);
	virtual ~Figur();

	/**
	 * Berechnet die Position der Figur fr das nchste Bild,
	 * setzt die nchste Animationsphase etc.
	 */
	void naechsterZustand();

	/**
	 * \returns die aktuelle Breite der Figur. FIXME: Ist die Breite
	 * des ersten Frames (der aktuellen Animationsphase oder berhaupt?)
	 * und das ist auch richtig, exakte Daten sind nur fr die Kollision
	 * wichtig und die erfolgt eh pixelgenau.
	 */
	int getBreite();
	
	/**
	 * \returns die aktuelle Hhe der Figur.
	 */
	int getHoehe();

	/**
	 * Testet, ob diese Figur mit einer anderen kollidiert.
	 * \param figur2 Die Figur, die auf Kollision geprft werden soll.
	 * \returns Wahr, falls eine Kollision vorliegt, sonst Falsch.
	 */
	Kollision pruefeKollision(const Figur& figur2);

	/*
	 * Zeichnet die Figur an ihrer aktuellen Position.
	 */
	void zeichnen();

	/**
	 * Setzt die Position der linken oberen Ecke der Figur.
	 * \param x Ist der X-Wert der linken oberen Ecke.
	 * \param y Ist der Y-Wert der linken oberen Ecke.
	 */
	void setzePosition(int x, int y);
	
	/**
	 * Setzt die X-Position der linken oberen Ecke der Figur.
	 * \param x Ist der X-Wert der linken oberen Ecke.
	 */
	void setX(int x);
	
	/**
	 * Setzt die Y-Position der linken oberen Ecke der Figur.
	 * \param y Ist der Y-Wert der linken oberen Ecke.
	 */
	void setY(int y);

	/**
	 * \returns die X-Position der linken oberen Ecke der Figur.
	 */
	int getX();
	
	/**
	 * \returns die X-Position der linken oberen Ecke der Figur.
	 */
	int getY();

	void steuerfunktionHinzufuegen(Steuerfunktion* steuerfunktion);

	void kollision(Kollision kollision);
	void kollision(Richtung richtung);
	
	void setzeBoundingBox(int bxl, int byt, int bxr, int byb);
	
	void setzeAbstaende(int abstandLinks, int abstandOben, int abstandUnten,
			int abstandRechts);
	

	int getDX();
	int getDY();
	void setDX(int dx);
	void setDY(int dy);
	
	void setVerschiebung(int verschiebung);
	
	bool faellt();
	bool isTot();
	void setTot(bool tot);
	
	/**
	 * \returns den Namen dieser Figur
	 */
	std::string getName();
	
	/**
	 * \returns den Typ dieser Figur
	 */
	FigurTyp getTyp();
	
	/**
	 * \returns ob diese Figur den Bildschirm nach unten verlassen,
	 * also "rausgefallen" ist.
	 */
	bool isUntenRausgefallen();
	
	void setRennt(bool rennt);
	
	void setzeUnverwundbar(int frames);
	bool isUnverwundbar();
	
	virtual std::string alsString();
	static Figur* ausString(SpriteManager* spriteManager, Level* level,
				std::string input_);

	static std::list<Figur*> ladeListeAusDatei(std::string dateiname,
				SpriteManager* spriteManager, Level* level);

	static void speichereListeAlsDatei(std::list<Figur*> objekte,
				std::string dateiname);

	// Fgt zustzliche Parameter in Stringform hinzu. Der Parameter wird
	// an eine interne Liste angehngt, die mitabgespeichert wird. Dann
	// wird versucht, den Parameter zu interpretieren.
	void setParameter(std::string parameter);

	unsigned int getSchaden();

	std::list<Figur*>& getBeruehrteKisten();

	int getPunkte();
};


#endif //FIGUR_H
