/*
	Copyright 2007 Zafir Mahmutovic, Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \author Zafir Mahmutovic
 * \author Martin Gernhard
 */

#include "level.h"

#include <fstream>
#include <iostream>
#include <sstream>

#include "logger.h"
#include "figur.h"
#include "utils.h"
#include "tile.h"

Level::Level( const std::string& dateiname, SpriteManager* spriteManager)
{
	this->tileManager = new TileManager(spriteManager);
	this->camera.x = 0;
	this->camera.y = 0;
	this->camera.w = SCREEN_WIDTH;
	this->camera.h = SCREEN_HEIGHT;

	//Tile Sprites initialisieren 
    if( laden( dateiname ) == false )
    {
        logger->error("Level " + dateiname + " konnte nicht geladen werden.");
        exit(1);
    }
    
    kameraAktualisieren();
}

Level::~Level()
{
	//Tiles freigeben
	for( int x = 0; x < anzahlTilesX; x++ )
	{
		for ( int y = 0; y < anzahlTilesY; y++ ){
			delete tiles[ x ][ y ];    
		}
	}
	
	delete tileManager;
}

bool Level::laden( const std::string& dateiname )
{
    //Die Datei mit der Level-Map ffnen
    std::ifstream map( dateiname.c_str() );
    
    if( map == NULL )
    {
        logger->error("Map-Datei '" + dateiname +  "' konnte nicht geladen werden");
        return false;    
    }
    
    map >> anzahlTilesX;
    map >> anzahlTilesY;
    
    levelBreite = anzahlTilesX * Tile::TILE_BREITE;
    levelHoehe  = anzahlTilesY * Tile::TILE_HOEHE;
    
    tiles.resize(anzahlTilesX);
    for (int x = 0; x < anzahlTilesX; x++){
    	tiles[x].resize(anzahlTilesY);
    }
    
    
    //Tiles initialisieren
    for( int y = 0; y < anzahlTilesY; y++ )
    {
    	for ( int x = 0; x < anzahlTilesX; x++)
    	{
	        //Legt fest welcher Tile Typ initialisiert wird
	        int typ = -1;
	    
	        //Tile aus der Map auslesen
	        map >> typ;
	        
	        //Falls es ein Problem gab beim Auslesen der Map
	        if( map.fail() == true )
	        {
	            logger->error("Fehler beim Lesen aus der Map-Datei '" + dateiname + "'");
	            return false;
	        }
	        
	        //Tile anlegen
	        const int xPos = x * Tile::TILE_BREITE;
	        const int yPos = y * Tile::TILE_HOEHE;
	        
	        if (typ >= TILETYP_LETZTES_TILE)
	        {
	        	// Tile-Typ konnte nicht erkannt werden.
	        	logger->error("Fehlerhafter Tile-Typ '" + itos(typ) + "' in " + dateiname + " an Stelle ("
	        		+ itos(x) + ", " + itos(y) + ")");
	            return false;
	        }
	        
			tiles[ x ][ y ] = tileManager->createTile(xPos, yPos, TileTyp(typ));
    	}
    }
    
    //Die Datei schliessen
    map.close();
    
    //Falls die Map korrekt geladen wurde.
    return true;
}

void Level::kameraAktualisieren()
{
    camera.y = 0;

    // Hlt die Kamera innerhalb der Grenzen
    camera.x = max(camera.x, 0);
    camera.x = min(camera.x, (levelBreite - camera.w));
//    camera.y = max(camera.y, 0);
//    camera.y = min(camera.y, (levelHoehe - camera.h));

	// Berechnet die Seiten der Kamera in Tile-Koordinaten    
    cameraLeft   = max(0, (camera.x / Tile::TILE_BREITE) );
    cameraTop    = max(0, (camera.y / Tile::TILE_HOEHE) );
    cameraRight  = min((anzahlTilesX - 1), ((camera.x + camera.w) / Tile::TILE_BREITE) );
    cameraBottom = min((anzahlTilesY - 1), ((camera.y + camera.h) / Tile::TILE_HOEHE) );
}

void Level::zeichnen()
{
    for( int x = cameraLeft; x <= cameraRight; x++ )
    {
    	for ( int y = cameraTop; y <= cameraBottom; y++ ){   
	        if(tiles[x][y] != NULL){
	            (tiles[x][y])->zeichnen( (tiles[x][y])->getX() - camera.x, (tiles[x][y])->getY() - camera.y);    
			}
    	}
    }	
}

bool Level::isEnde()
{
	return !(camera.x < levelBreite - SCREEN_WIDTH);	
}

void Level::setVerschiebung( int verschiebung )
{
	camera.x = verschiebung;
	kameraAktualisieren();
}

bool Level::isSolid(int x, int y)
{

	if (isInvalidTile(x, y)){		// Ist ausserhalb des Levels.
		return false;
	}else if ((tiles[x][y] == NULL) ){
		return false;
	}else{
		return (tiles[x][y])->isSolid();
	}
}

TileTyp Level::getTileTypAnPosition(int x, int y)
{
	return getTileTyp(x / Tile::TILE_BREITE, y / Tile::TILE_HOEHE);
}

TileTyp Level::getTileTyp(int tileX, int tileY)
{
	if (isInvalidTile(tileX, tileY)){
		return TILETYP_LEER;
	}
	if (tiles[tileX][tileY] == NULL){
		return TILETYP_LEER;
	}
	return (tiles[tileX][tileY])->getTileTyp();
}

void Level::setTileTypAnPosition(int x, int y, TileTyp typ)
{
	setTileTyp(x / Tile::TILE_BREITE, y / Tile::TILE_HOEHE, typ);
}

void Level::setTileTyp(int tileX, int tileY, TileTyp typ)
{
	if (isInvalidTile(tileX, tileY)){
		return;
	}else{
		delete tiles[tileX][tileY];
		tiles[tileX][tileY] = tileManager->createTile(tileX * Tile::TILE_BREITE,
					tileY * Tile::TILE_HOEHE, typ);
	}
}


bool Level::isInvalidTile(int x, int y)
{
	if ( (x < 0) || (x >= anzahlTilesX) || (y < 0) || (y >= anzahlTilesY) ){
		return true;
	}else{
		return false;
	} 	
}

void Level::speichern(std::string dateiname)
{
	std::ofstream mapdatei( dateiname.c_str() );

	if( mapdatei == NULL )
	{
		logger->error("Map-Datei '" + dateiname +  "' konnte nicht geschrieben werden.");
	}
	
	mapdatei << anzahlTilesX << std::endl;
	mapdatei << anzahlTilesY << std::endl;
	
	for( int y = 0; y < anzahlTilesY; y++ )
	{
		for ( int x = 0; x < anzahlTilesX; x++ ){
			mapdatei << getTileTyp(x, y) << " ";
		}
		
		mapdatei << std::endl;
	}
	
	mapdatei.close();
}

void Level::groesseAendern(int anzahlTilesX, int anzahlTilesY){

	int altAnzahlTilesX = this->anzahlTilesX;
	int altAnzahlTilesY = this->anzahlTilesY;

	this->anzahlTilesX = anzahlTilesX;
	this->anzahlTilesY = anzahlTilesY;

	levelBreite = anzahlTilesX * Tile::TILE_BREITE;
	levelHoehe  = anzahlTilesY * Tile::TILE_HOEHE;

	// Falls der Level kleiner geworden ist, berflssige Tiles entfernen.
    for ( int x = anzahlTilesX; x < altAnzahlTilesX; x++ ){
    	for ( int y = anzahlTilesY; y < altAnzahlTilesY; y++ ){
    		delete tiles[x][y];
    	}
    }

   	tiles.resize(anzahlTilesX);

	// Neue Gre setzen und neu entstandene Tile mit NULL initialisieren.
    for (int x = 0; x < anzahlTilesX; x++){
    	tiles[x].resize(anzahlTilesY, NULL);
    }
}

void Level::kisteHinzufuegen(Kiste* kiste)
{
	this->kisten.push_back(kiste);
}

Kiste* Level::sucheKisteAnPosition(int tileX, int tileY)
{
	const int xPos = tileX * Tile::TILE_BREITE;
	const int yPos = tileY * Tile::TILE_HOEHE;


	for( std::list<Kiste*>::iterator iter = kisten.begin(); iter != kisten.end(); iter++){
		Kiste* kiste = *iter;
		if (	(kiste->getX() <= xPos && xPos <= (kiste->getX() + kiste->getBreite() - 1)) &&
				(kiste->getY() <= yPos && yPos <= (kiste->getY() + kiste->getHoehe() - 1)) )
		{
			// TODO: Zur "Liste der berhrten Kisten" hinzufgen.
			// TODO: Kann man hier auch feststellen, ob oben oder unten berhrt?
			//			evtl. xPos und yPos speichern, die ja die
			//			"versuchsweise" Position darstellen.
			return kiste;
		}
	}

	return NULL;
}

void Level::kisteEntfernen(Kiste* kiste)
{
	kisten.remove(kiste);
}

void Level::neustart()
{
	kisten.clear();
}
