/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * \author Martin Gernhard
 */

#include <stdlib.h>
#include <iostream>
#include <vector>
#include <utility>

#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>
#include <SDL_mixer.h>

#include "logger.h"
#include "utils.h"
#include "sprite.h"
#include "figur.h"
#include "ring.h"
#include "stachel.h"
#include "kiste.h"
#include "herzkiste.h"

#include "level.h"
#include "tile.h"
#include "tilemanager.h"
#include "bitmapfont.h"

#include "linksrechtsbewegung.h"
#include "gegnersprung.h"

#include "kollision.h"

enum GegnerTyp {GEGNERTYP_KIRSCHEN, GEGNERTYP_KUERBIS, GEGNERTYP_STACHEL,
			GEGNERTYP_RING,
			GEGNERTYP_GELDKISTE,
			GEGNERTYP_GELDKISTE_GROSS,
			GEGNERTYP_HERZKISTE,
			GEGNERTYP_SWORDSMAN,
			GEGNERTYP_TITAN, GEGNERTYP_SMALL_ONI, GEGNERTYP_BIG_ONI,
			GEGNERTYP_SPIRIT,
			GEGNERTYP_LETZTER_GEGNER};

class GegnerVorlage
{
private:
	Sprite* bild;
	std::string animationsdatei;
public:
	// Notwendig, um diesen Typ in einer Map verwenden zu knnen.
	GegnerVorlage()
	{
		bild = NULL;
		animationsdatei = "";
	}
	GegnerVorlage(Sprite* bild, std::string animationsdatei)
	{
		this->bild = bild;
		this->animationsdatei = animationsdatei;
	}
	Sprite* getBild()
	{
		return bild;
	}

	std::string getAnimationsdatei()
	{
		return animationsdatei;
	}
};

std::map<GegnerTyp, GegnerVorlage> gegnerVorlagen;

void initGegnerVorlagen(SpriteManager* spriteManager)
{
	gegnerVorlagen[GEGNERTYP_KIRSCHEN] = GegnerVorlage(
				spriteManager->createSprite("sprites/cherries.png"),
				"sprites/cherries.anim");
	gegnerVorlagen[GEGNERTYP_KUERBIS]  = GegnerVorlage(
				spriteManager->createSprite("sprites/pumpkin.png"),
				"sprites/pumpkin.anim");
	gegnerVorlagen[GEGNERTYP_STACHEL]  = GegnerVorlage(
				spriteManager->createSprite("sprites/stachel1.png"),
				"sprites/stachel.anim");
	gegnerVorlagen[GEGNERTYP_RING]     = GegnerVorlage(
				spriteManager->createSprite("sprites/ring1.png"),
				"sprites/ring.anim");
	gegnerVorlagen[GEGNERTYP_GELDKISTE]     = GegnerVorlage(
				spriteManager->createSprite("sprites/fragezeichen2.png"),
				"sprites/kiste.anim");
	gegnerVorlagen[GEGNERTYP_GELDKISTE_GROSS]     = GegnerVorlage(
				spriteManager->createSprite("sprites/kiste_gross.png"),
				"sprites/kiste_gross.anim");
	gegnerVorlagen[GEGNERTYP_HERZKISTE]     = GegnerVorlage(
				spriteManager->createSprite("sprites/herzkiste.png"),
				"sprites/kiste.anim");
	gegnerVorlagen[GEGNERTYP_SWORDSMAN]= GegnerVorlage(
				spriteManager->createSprite("sprites/sword_gehen1.png"),
				"sprites/sword.anim");
	gegnerVorlagen[GEGNERTYP_TITAN]    = GegnerVorlage(
				spriteManager->createSprite("sprites/titan_gehen1.png"),
				"sprites/titan.anim");
	gegnerVorlagen[GEGNERTYP_SMALL_ONI]= GegnerVorlage(
				spriteManager->createSprite("sprites/soni_sprung2.png"),
				"sprites/soni.anim");
	gegnerVorlagen[GEGNERTYP_BIG_ONI]  = GegnerVorlage(
				spriteManager->createSprite("sprites/boni_sprung2.png"),
				"sprites/boni.anim");
	gegnerVorlagen[GEGNERTYP_SPIRIT]   = GegnerVorlage(
				spriteManager->createSprite("sprites/spirit1.png"),
				"sprites/spirit.anim");
}

void objektHinzufuegen(GegnerTyp gegnerTyp, std::list<Figur*>& gegner,
			Level* level, SpriteManager* spriteManager, int objektX,
			int objektY)
{
	Figur* figur;
	switch(gegnerTyp){
		case GEGNERTYP_KIRSCHEN:
		case GEGNERTYP_SWORDSMAN:
			figur = new Figur("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			figur->steuerfunktionHinzufuegen(new LinksrechtsBewegung(figur));
			break;
		case GEGNERTYP_TITAN:
		{
			figur = new Figur("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			LinksrechtsBewegung* bewegung = new LinksrechtsBewegung(figur);
			bewegung->setGeschwindigkeit(3);
			figur->steuerfunktionHinzufuegen(bewegung);
			figur->setParameter("Schaden 2");
		}
			break;
		case GEGNERTYP_KUERBIS:
		case GEGNERTYP_BIG_ONI:
			figur = new Figur("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			figur->steuerfunktionHinzufuegen(new LinksrechtsBewegung(figur));
			figur->steuerfunktionHinzufuegen(new Gegnersprung(figur));
			figur->setParameter("Schaden 2");
			break;
		case GEGNERTYP_SMALL_ONI:
			figur = new Figur("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			figur->steuerfunktionHinzufuegen(new Gegnersprung(figur));
			break;
		case GEGNERTYP_STACHEL:
			figur = new Stachel("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			break;
		case GEGNERTYP_SPIRIT:
			figur = new Figur("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			figur->setParameter("Schaden 4");
			figur->setParameter("Gravitation 0");
			break;
		case GEGNERTYP_RING:
			figur = new Ring("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			break;
		case GEGNERTYP_GELDKISTE:
		case GEGNERTYP_GELDKISTE_GROSS:
			figur = new Kiste("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			break;
		case GEGNERTYP_HERZKISTE:
			figur = new HerzKiste("Auto-Generated-Object",
						gegnerVorlagen[gegnerTyp].getAnimationsdatei(), level,
						spriteManager);
			break;
		default:
			break;
	}
	figur->setzePosition(objektX, objektY);

	gegner.push_back(figur);
}

Figur* findeObjektUnterCursor(std::list<Figur*>& gegner, Figur* cursor)
{
	for (std::list<Figur*>::iterator iter = gegner.begin(); iter != gegner.end(); iter++){
		Figur* g = *iter;
		Kollision kollision = cursor->pruefeKollision(*g);
		if (kollision.kollidiert()){
			logger->debug("Gefunden");
			return g;
		}		
	}

	return NULL;
}

GegnerTyp& operator++(GegnerTyp& gegnerTyp)
{
	unsigned int tmp = gegnerTyp;

	tmp++;
	if (tmp == GEGNERTYP_LETZTER_GEGNER){
		tmp = 0;
	}
	gegnerTyp = GegnerTyp(tmp);
	return gegnerTyp;
}

GegnerTyp& operator--(GegnerTyp& gegnerTyp)
{
	unsigned int tmp = gegnerTyp;

	if (tmp == 0){
		tmp = GEGNERTYP_LETZTER_GEGNER - 1;
	}else{
		tmp--;
	}
	gegnerTyp = GegnerTyp(tmp);
	return gegnerTyp;
}

int main(int argc, char **argv)
{
    Logger::initLogging(Logger::LOGLEVEL_WARN);
	SDL_OpengGL_Init();

	const bool objekteAnimieren = false;
	
	// SpriteManager initialisieren
	SpriteManager spriteManager;

	// TileManager initialisieren
	TileManager tileManager(&spriteManager);

	// Fonts laden
	BitmapFont weiss(&spriteManager, "fonts/bitstreamverasanswhite.png");

	// Kommandozeilenparameter auswerten
	std::string levelName = "level/level1";
	if (argc > 1){
		levelName = std::string(argv[1]);
	}

	// Leveldatei laden
	Level* level = new Level( levelName + ".map", &spriteManager );

	if (argc == 4){
		// Levelgre wurde bergeben, Gre des Levels ndern.
		int anzahlTilesX = atoi(argv[2]);
		int anzahlTilesY = atoi(argv[3]);
		level->groesseAendern(anzahlTilesX, anzahlTilesY);
	}

	// Gegner initialisieren
	std::list<Figur*> gegner = Figur::ladeListeAusDatei(levelName + ".objects", &spriteManager, level);

	initGegnerVorlagen(&spriteManager);

	Figur* cursor = new Figur("Cursor", "sprites/cursor.anim", level, &spriteManager);
	Figur* auswahl = NULL;

    //Title Bar
	SDL_WM_SetCaption("Leveleditor", "Leveleditor");

	SDL_ShowCursor(0);

	// Hauptschleife, fragt die Tastatur ab und ndert ggf. die
	// Bewegungsrichtung. Fgt beim Drcken der Leertaste Sprungfunktionen
	// zur Spielerfigur hinzu.
    // bool gehtRechts = false;

	bool running = true;
	int verschiebung = 0;
	TileTyp typ = TILETYP_BLOCK;
	GegnerTyp gegnerTyp = GegnerTyp(0);

	bool tileModus = true;
	bool gitter = false;
	
	while(running) {
		int startticks = SDL_GetTicks();

		SDL_PumpEvents();

		Uint8* keystates = SDL_GetKeyState(NULL);
		if (keystates[SDLK_RIGHT] || keystates[SDLK_d]){
			verschiebung += Tile::TILE_BREITE;
		}else if (keystates[SDLK_LEFT] || keystates[SDLK_a]){
			verschiebung -= Tile::TILE_BREITE;
		}

		verschiebung = max(verschiebung, 0);
		verschiebung = min(verschiebung, (level->getBreite() - SCREEN_WIDTH));

		int mausX, mausY;
		if (SDL_GetMouseState(&mausX, &mausY) & SDL_BUTTON(1)){
			if (tileModus && (auswahl == NULL)){
				level->setTileTypAnPosition(mausX + verschiebung, mausY, typ);
			}
		}else if (SDL_GetMouseState(&mausX, &mausY) & SDL_BUTTON(3)){
			if (tileModus && (auswahl == NULL)){
				level->setTileTypAnPosition(mausX + verschiebung, mausY, TILETYP_LEER);
			}
		}

		int gitterX = mausX - (mausX % Tile::TILE_BREITE);
		int gitterY = mausY - (mausY % Tile::TILE_HOEHE); 

		if (gitter){
			mausX = gitterX;
			mausY = gitterY;
		}

		// Tasten, bei denen Drcken und Loslassen als gretrennte Ereignisse
		// behandelt werden.
		SDL_Event event;
		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_KEYDOWN:
				switch( event.key.keysym.sym ){
					case SDLK_q:
						running = false;
						break;
					case SDLK_PLUS:
					case SDLK_KP_PLUS:
						level->groesseAendern((level->getBreite() / Tile::TILE_BREITE) + 1,
									level->getHoehe() / Tile::TILE_HOEHE);
						break;
					case SDLK_MINUS:
					case SDLK_KP_MINUS:
						if (level->getBreite() > SCREEN_WIDTH){
							level->groesseAendern((level->getBreite() / Tile::TILE_BREITE) - 1,
										level->getHoehe() / Tile::TILE_HOEHE);
						}
						break;
					case SDLK_m:
						tileModus = !tileModus;
						break;
					case SDLK_DELETE:
						if (auswahl != NULL){
							gegner.remove(auswahl);
							auswahl = NULL;
						}
						break;
					case SDLK_g:
						gitter = !gitter;
						break;
					case SDLK_w:
					case SDLK_UP:
						if (tileModus){
							++typ;
						}else{
							++gegnerTyp;
						}
						break;
					case SDLK_s:
					case SDLK_DOWN:
						if (tileModus){
							--typ;
						}else{
							--gegnerTyp;
						}
						break;
					default:
						break;
				}
				break;
			case SDL_KEYUP:
				switch( event.key.keysym.sym ){
                    default:
						break;
				}
				break;
			case SDL_QUIT:
				running = false;
				break;
			case SDL_MOUSEBUTTONDOWN:
				switch(event.button.button){
					case SDL_BUTTON_LEFT:
						if (auswahl == NULL && !tileModus){
							objektHinzufuegen(gegnerTyp, gegner, level,
									&spriteManager, mausX + verschiebung, mausY);
						}
						break;
					case SDL_BUTTON_MIDDLE:
						if (auswahl != NULL){
							auswahl = NULL;
						}else{
							auswahl = findeObjektUnterCursor(gegner, cursor);
						}
						break;
					case SDL_BUTTON_RIGHT:
					{
						if (!tileModus && (auswahl == NULL)){
							Figur* kandidat = findeObjektUnterCursor(gegner, cursor);
							if (kandidat != NULL){
								gegner.remove(kandidat);
							}
						}else if (auswahl != NULL){
							gegner.remove(auswahl);
							auswahl = NULL;
						}
					}
						break;
					case SDL_BUTTON_WHEELUP:
						if (tileModus){
							++typ;
						}else{
							++gegnerTyp;
						}
						break;
					case SDL_BUTTON_WHEELDOWN:
						if (tileModus){
							--typ;
						}else{
							--gegnerTyp;
						}
						break;
				}
				break;
			}
		}
		
		level->setVerschiebung( verschiebung );
		
        /* Das Bild neu aufbauen. */

		ClearScreen();

        level->zeichnen();

		if (auswahl != NULL){
			auswahl->setzePosition(mausX + verschiebung, mausY);
		}

		for ( std::list<Figur*>::iterator iter = gegner.begin(); iter != gegner.end(); iter++){
			Figur* g = *iter;
			if (objekteAnimieren){
				g->naechsterZustand();
			}
			g->setVerschiebung( verschiebung );
			g->zeichnen();
			zeichneHohlRechteck(g->getX() - verschiebung, g->getY(), g->getBreite(), g->getHoehe());
		}


		/* HUD */
		Tile* tmpTile = tileManager.createTile(0, 480, typ);
		tmpTile->zeichnen();
		if (tileModus && auswahl == NULL){
			tmpTile->zeichnen(gitterX, gitterY);
		}
		delete tmpTile;

		if (!tileModus && (auswahl == NULL) ){
			Sprite* sprite = gegnerVorlagen[gegnerTyp].getBild();
			sprite->zeichnen(mausX, mausY);
			zeichneHohlRechteck(mausX, mausY, sprite->getBreite(), sprite->getHoehe()); 
		}
		
		weiss.schreibeZeile("m: Modus, g: Gitter", 1 * Tile::TILE_BREITE, SCREEN_HEIGHT);


		if (gitter){
			for (int x = 1; x < (SCREEN_WIDTH / Tile::TILE_BREITE); x++){
				zeichneRechteck(x * Tile::TILE_BREITE, 0, 1, SCREEN_HEIGHT);
			}

			for (int y = 1; y < (SCREEN_HEIGHT / Tile::TILE_HOEHE); y++){
				zeichneRechteck(0, y * Tile::TILE_HOEHE, SCREEN_WIDTH, 1);
			}
		}

		cursor->setzePosition(mausX + verschiebung, mausY);
		cursor->setVerschiebung( verschiebung );
		if ( auswahl == NULL && tileModus){
			cursor->zeichnen();
		}

		/* Das Bild ist jetzt aufgebaut. */
         

		// Pageflip: Back- und Frontbuffer werden vertauscht
		SwapBuffers();

		// Konstante Bildrate von 50fps.
		int ticksleft = (1000/50) - (SDL_GetTicks() - startticks);
		if (ticksleft > 0){ SDL_Delay(ticksleft); }
	}



	// Level speichern
	level->speichern( levelName + ".map" );
	Figur::speichereListeAlsDatei( gegner, levelName + ".objects" );

	// Aufrumen.
	delete level;
	delete cursor;

	// Verbliebene Figuren lschen.
	for (std::list<Figur*>::iterator iter = gegner.begin(); iter != gegner.end();){
		Figur* g = *iter;
		iter = gegner.erase(iter);
		delete g;			
	}

	    Logger::shutdownLogging();
	return 0;
}
