/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Einfache Klasse, die eine Bilddatei in eine passende OpenGL-Textur ldt
 * und Befehle bereitstellt, um diese auf dem Bildschirm zu zeichnen.
 * \author Martin Gernhard
 */

#include <iostream>
#include <string>

#include <SDL_image.h>

#include "sprite.h"
#include "utils.h"
#include "logger.h"

Sprite::Sprite(const std::string& dateiname)
{
	ladeSurface(dateiname);

#ifdef OUTPUT_SDL
	erzeugeSpiegelBilddaten();
#else
	erzeugeTexturAusSurface();
#endif
}

Sprite::Sprite(	SDL_Surface* bilddaten )
{
	this->bilddaten = bilddaten;
	this->breite = bilddaten->w;
	this->hoehe = bilddaten->h;
	
#ifdef OUTPUT_SDL
	erzeugeSpiegelBilddaten();
#else
	erzeugeTexturAusSurface();
#endif
}

Sprite::~Sprite()
{
#ifndef OUTPUT_SDL
	glDeleteTextures(1, &textur);
#else
	SDL_FreeSurface(spiegelBilddaten);
#endif
	SDL_FreeSurface(bilddaten);
	logger->debug("Sprite entfernt.");
}

void Sprite::gespiegeltZeichnen(int x, int y, int breite)
{
	const int verschiebung = (texb - breite);
#ifndef OUTPUT_SDL
	GL_TexEnable(1);
	GL_Bind(textur);

	glBegin(GL_QUADS);
	glTexCoord2i(0, 0);	glVertex2i((x - verschiebung) + texb, y);
	glTexCoord2i(1, 0);	glVertex2i(x - verschiebung,        y);
	glTexCoord2i(1, 1);	glVertex2i(x - verschiebung,        y + texh);
	glTexCoord2i(0, 1);	glVertex2i((x - verschiebung) + texb, y + texh);
	glEnd();
#else
	SDL_Rect ziel;
	ziel.x = x - verschiebung;
	ziel.y = y;

	SDL_BlitSurface(spiegelBilddaten, NULL, mainScreen, &ziel);
#endif
}

#ifdef OUTPUT_SDL
void Sprite::erzeugeSpiegelBilddaten()
{
	// Aufrunden der Seitenlngen auf Zweierpotenzen.
	texb = naechsteZweierpotenz(breite);
	texh = naechsteZweierpotenz(hoehe);

	spiegelBilddaten = SDL_CreateRGBSurface(SDL_SWSURFACE, texb, texh, 32,
				rmask, gmask, bmask, amask);

	SDL_LockSurface(bilddaten);
	SDL_LockSurface(spiegelBilddaten);

	for (int y = 0; y < texh; y++){
		for (int x = 0; x < texb; x++){
			int position1 = x + y * bilddaten->w;
			int position2 = ((texb - 1) - x) + y * bilddaten->w;

			Uint32 pixel1 = *((Uint32*)bilddaten->pixels + position1);
			*((Uint32*)spiegelBilddaten->pixels + position2) = pixel1;
		}
	}


	SDL_UnlockSurface(bilddaten);
	SDL_UnlockSurface(spiegelBilddaten);
}
#endif

void Sprite::gespiegeltZeichnen(int x, int y)
{
	gespiegeltZeichnen(x, y, this->breite);
}

void Sprite::zeichnen(int x, int y)
{
#ifndef OUTPUT_SDL
	GL_TexEnable(1);
	GL_Bind(textur);

	glBegin(GL_QUADS);
	glTexCoord2i(0, 0);	glVertex2i(x,        y);
	glTexCoord2i(1, 0);	glVertex2i(x + texb, y);
	glTexCoord2i(1, 1);	glVertex2i(x + texb, y + texh);
	glTexCoord2i(0, 1);	glVertex2i(x,        y + texh);
	glEnd();
#else
	SDL_Rect ziel;
	ziel.x = x;
	ziel.y = y;

	SDL_BlitSurface(bilddaten, NULL, mainScreen, &ziel);
#endif
}

#ifndef OUTPUT_SDL
void Sprite::zeichnen(int x, int y, int breite, int hoehe)
{
	GL_TexEnable(1);
	GL_Bind(textur);


	const int echteBreite = int( ((double)breite/this->breite) * texb );
	const int echteHoehe  = int( ((double)hoehe /this->hoehe ) * texh );

	glBegin(GL_QUADS);
	glTexCoord2i(0, 0);	glVertex2i(x,        y);
	glTexCoord2i(1, 0);	glVertex2i(x + echteBreite, y);
	glTexCoord2i(1, 1);	glVertex2i(x + echteBreite, y + echteHoehe);
	glTexCoord2i(0, 1);	glVertex2i(x,               y + echteHoehe);
	glEnd();
}
#endif

int Sprite::getBreite()
{
	return breite;
}

int Sprite::getHoehe()
{
	return hoehe;
}

SDL_Surface* Sprite::getBilddaten(){
	return bilddaten;
}


/**
 * Ldt eine 32bit RGBA Textur aus einer Bilddatei. Ein evtl. vorhandener
 * Alphakanal wird dabei erhalten, ansonsten werden die Alphawerte der Textur
 * durchgehend auf undurchsichtig gesetzt.
 * Die Seitenlngen der Grafik werden bei der Umwandlung auf die nchsthheren
 * Zweierpotenzen aufgerundet, es handelt sich dabei um eine Einschrnkung von
 * OpenGL selbst.
 *
 * Breite und Hhe der ursprnglichen Grafik werden in grafb und grafh gespeichert.
 * Breite und Hhe der Textur werden in texb und texh gespeichert.
 * Wurde nur mit PNG getestet.
 */
void Sprite::ladeSurface(const std::string& dateiname)
{
	// Die Originaltextur, wie sie aus der Datei geladen wurde.
	SDL_Surface* image;
	// Die in ein fr OpenGL "verdauliches" Format umgewandelte Textur.
	SDL_Surface* conv;
	
	// Laden der Bilddatei.
	image = IMG_Load(dateiname.c_str());
	if (!image) {
		logger->error(dateiname + " konnte nicht geladen werden.");
		logger->error(SDL_GetError());
		exit(1);
	}

	breite = image->w;
	hoehe = image->h;

	// Aufrunden der Seitenlngen auf Zweierpotenzen.
	texb = naechsteZweierpotenz(breite);
	texh = naechsteZweierpotenz(hoehe);


	// Formatumwandlung durchfhren, falls erforderlich. Abbrechen,
	// falls das Format nicht untersttzt wird.
	switch(image->format->BitsPerPixel){
	case 32:
	case 24:
		conv = SDL_CreateRGBSurface(SDL_SWSURFACE, texb, texh, 32,
					rmask, gmask, bmask, amask);
		// 24bit-Grafiken wird ein komplett undurchsichtiger Alphakanal zugewiesen.
		// 32bit-Grafiken werden inklusive Alphakanal kopiert.
		// Siehe dazu Dokumentation zu SDL_SetAlpha.
		SDL_SetAlpha(image, 0, SDL_ALPHA_TRANSPARENT);

		// Kopieren auf die neue Surface und Lschen der alten Surface.
		SDL_BlitSurface(image, 0, conv, 0);
		SDL_FreeSurface(image);
		image = conv;
		bilddaten = conv;
		break;
	default:
		logger->error("Nicht untersttztes Bildformat: " + itos((int)(image->format->BitsPerPixel))
					+ "bpp. (" + dateiname + ")");
		exit(1);
		break;
	}


	// Die erzeugte Surface wird nicht gelscht, denn sie ist Teil der
	// Rckgabewerte
	// SDL_FreeSurface(image);
}

#ifndef OUTPUT_SDL
void Sprite::erzeugeTexturAusSurface( ){

	int grafb = bilddaten->w;
	int grafh = bilddaten->h;

	// Aufrunden der Seitenlngen auf Zweierpotenzen.
	texb = naechsteZweierpotenz(grafb);
	texh = naechsteZweierpotenz(grafh);

	/* Erstellen der OpenGL-Textur */
	// Anzahl der Bytes pro Zeile automatisch bestimmen lassen.
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);

	// Leere Textur erstellen.
	glGenTextures(1, &textur);
	// Die soeben erzeugte Textur als aktuelle Textur setzen.
	glBindTexture(GL_TEXTURE_2D, textur);

	// Filter fr die Verkleinerung / Vergrerung der Textur setzen.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// Textur aus der SDL-Surface erzeugen.
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, texb, texh, 0, GL_RGBA,
		GL_UNSIGNED_BYTE, bilddaten->pixels);

	// Die erzeugte Surface wird nicht gelscht, denn sie ist Teil des
	// Sprites.
	// SDL_FreeSurface(image);

}
#endif
