/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \author Martin Gernhard
 */

#include <iostream>

#include "logger.h"
#include "spritemanager.h"
#include <SDL_image.h>
#include "utils.h"
#include <sstream>	

SpriteManager::SpriteManager()
{
}

SpriteManager::~SpriteManager()
{
	for (std::map<std::string, Sprite*>::iterator iter = sprites.begin(); iter != sprites.end(); iter++){
		Sprite* sprite = (*iter).second;
		delete sprite;
	}
	sprites.clear();
}

void SpriteManager::spritesAuflisten()
{
	logger->debug("<SpriteManager>");
	for (std::map<std::string, Sprite*>::iterator iter = sprites.begin(); iter != sprites.end(); iter++){
		std::string dateiname = (*iter).first;
		Sprite* sprite = (*iter).second;
		logger->debug(dateiname + "|" + itos(sprite->texb) + "|" + itos(sprite->texh)
					+ "|" + itos(sprite->breite) + "|" + itos(sprite->hoehe));
	}
	sprites.clear();
	logger->debug("</SpriteManager>");
}


Sprite* SpriteManager::createSprite(const std::string dateiname)
{
	Sprite* sprite = NULL;
	
	if (sprites.count(dateiname) > 0){
		// Sprite ist bereits geladen, Zeiger darauf zurckgeben. 
		sprite = sprites[dateiname];
	}else{
		// Sprite ist noch nicht geladen, also laden wir es mal.
		logger->debug("Sprite " + dateiname + " wird geladen.");
		sprite = new Sprite(dateiname);
		sprites[dateiname] = sprite;
	}

	
	return sprite;
}

std::vector<Sprite*> SpriteManager::erzeugeFont(const std::string dateiname,
			int& zeichenBreite, int& zeichenHoehe)
{
	std::vector<Sprite*> fontVector;

	// Die Originaltextur, wie sie aus der Datei geladen wurde.
	SDL_Surface* image;
	// Die in ein fr OpenGL "verdauliches" Format umgewandelte Textur.
	SDL_Surface* conv;
	
	// Laden der Bilddatei.
	image = IMG_Load(dateiname.c_str());
	if (!image) {
		logger->error(dateiname + " konnte nicht geladen werden.");
		logger->error(SDL_GetError());
		exit(1);
	}

	int grafb = image->w;
	int grafh = image->h;

	// Aufrunden der Seitenlngen auf Zweierpotenzen.
	int texb = naechsteZweierpotenz(grafb);
	int texh = naechsteZweierpotenz(grafh);

	int bildBreite = image->w;
	int bildHoehe = image->h;

	// Formatumwandlung durchfhren, falls erforderlich. Abbrechen,
	// falls das Format nicht untersttzt wird.
	switch(image->format->BitsPerPixel){
	case 32:
	case 24:
		conv = SDL_CreateRGBSurface(SDL_SWSURFACE, texb, texh, 32,
					rmask, gmask, bmask, amask);
		// 24bit-Grafiken wird ein komplett undurchsichtiger Alphakanal zugewiesen.
		// 32bit-Grafiken werden inklusive Alphakanal kopiert.
		// Siehe dazu Dokumentation zu SDL_SetAlpha.
		SDL_SetAlpha(image, 0, SDL_ALPHA_TRANSPARENT);

		// Kopieren auf die neue Surface und Lschen der alten Surface.
		SDL_BlitSurface(image, 0, conv, 0);
		SDL_FreeSurface(image);
		image = conv;
		break;
	default:
		logger->error("Nicht untersttztes Bildformat: " + itos((int)(image->format->BitsPerPixel))
					+ "bpp. (" + dateiname + ")");
		exit(1);
		break;
	}

	int zeichenNummer = 0;
	
	zeichenBreite = bildBreite / 16;
	zeichenHoehe = bildHoehe / 16;
	
	// Aufrunden der Seitenlngen auf Zweierpotenzen.
	int zeichenTexturBreite = naechsteZweierpotenz(zeichenBreite);
	int zeichenTexturHoehe = naechsteZweierpotenz(zeichenHoehe);
	
	for (int y = 0; y < 16; y++){
		for (int x = 0; x < 16; x++){
			std::stringstream pseudoDateiname;
			pseudoDateiname << dateiname << '-' << zeichenNummer;
			
			// Prfen, ob dieser Ausschnitt bereits im SpriteManager ist.
			if (sprites.count(pseudoDateiname.str()) > 0){
				fontVector.push_back(sprites[pseudoDateiname.str()]);
			}else{
				SDL_Surface* zeichenSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, zeichenTexturBreite, zeichenTexturHoehe, 32,
							rmask, gmask, bmask, amask);
				SDL_SetAlpha(image, 0, SDL_ALPHA_TRANSPARENT);
		
				// Kopieren auf die neue Surface.
				SDL_Rect src;
				src.x = x * zeichenBreite;
				src.y = y * zeichenHoehe;
				src.w = zeichenBreite;
				src.h = zeichenHoehe;
				SDL_BlitSurface(image, &src, zeichenSurface, 0);
	
				// Neues Sprite aus der Surface erzeugen und zum Rckgabe-Vektor
				// und zur Map des Managers hinzufgen.		
				Sprite* zeichen = new Sprite( zeichenSurface );
				fontVector.push_back(zeichen);
				sprites[pseudoDateiname.str()] = zeichen;
			}
			zeichenNummer++;
		}
	}

	SDL_FreeSurface(image);

	return fontVector;
}

#ifdef OUTPUT_SDL
Sprite* SpriteManager::erzeugeSDLHintergrund(const std::string dateiname)
{
	Sprite* sprite = new Sprite(dateiname);
	SDL_FreeSurface ( sprite->bilddaten );
	SDL_Surface* tmp = IMG_Load(dateiname.c_str());
	sprite->bilddaten = SDL_DisplayFormat( tmp );
	SDL_FreeSurface ( tmp );
	return sprite;
}
#endif
