/*
	Copyright 2007 Martin Gernhard

	This file is part of Jump And Run

    Jump And Run is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Hilfsfunktionen.
 * \author Martin Gernhard
 */

#include <SDL.h>

#ifndef OUTPUT_SDL
	#include <SDL_opengl.h>
#endif

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>

#include "utils.h"
#include "logger.h"
#include <SDL_mixer.h>

/*
 * Berechnet die nchsthhere Zweierpotenz zu einer gegebenen, positiven,
 * ganzen Zahl.
 */
int naechsteZweierpotenz(int num)
{
	int ergebnis = 2;

	while(num > ergebnis){
		ergebnis = ergebnis * 2;
	}

	return ergebnis;
}

/**
 * Gibt eine benutzerdefinierte Fehlermeldung sowie die letzte
 * SDL-Fehlermeldung aus.
 */
void sdlFehlerAusgeben(const std::string &meldung)
{
	logger->error(meldung);
	logger->error("Letzte SDL-Fehlermeldung: ");
	logger->error(SDL_GetError());
}

#ifndef OUTPUT_SDL
/**
 * Aktiviert oder deaktiviert das Texturieren.
 * Diese Funktion ist ein "wrapper" fr glEnable(GL_TEXTURE_2D) bzw.
 * glDisable(GL_TEXTURE_2D) und ruft diese Funktionen nur dann auf,
 * wenn der Zustand tatschlich gendert werden soll, da der Wechsel
 * zeitaufwndig ist.
 */
void GL_TexEnable(int enable)
{
	static int gl_state_texenable = 0;

	if (gl_state_texenable == enable)
		return;

	if (enable)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);

	gl_state_texenable = enable;
}

/**
 * Setzt eine Textur als die aktuell verwendete Textur.
 * Diese Funktion ist ein "wrapper" fr glBindTexture(GL_TEXTURE_2D, texture);
 * und ruft diese Funktion nur dann auf, wenn tatschlich eine andere Textur
 * verwendet werden soll, da der Wechsel zeitaufwndig ist.
 */
void GL_Bind(GLuint texture)
{
	static GLuint gl_state_texture = 0;

	if (gl_state_texture == texture)
		return;

	glBindTexture(GL_TEXTURE_2D, texture);

	gl_state_texture = texture;
}
#endif


// Globale SDL-Surface
SDL_Surface* mainScreen;

SDL_Surface* SDL_OpengGL_Init()
{
	SDL_Surface *screen = NULL;

	/* SDL initialisieren */
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		sdlFehlerAusgeben("SDL-Video konnte nicht initialisiert werden.");
		exit(1);
	}

	// Aufrumen, falls das Programm unerwartet beendet wird.
	atexit(SDL_Quit);

	if (SDL_Init(SDL_INIT_AUDIO) != 0){
		sdlFehlerAusgeben("Warnung: SDL-Audio konnte nicht initialisiert werden.");
	}

	if (Mix_OpenAudio(44100, AUDIO_S16, 2, 512) < 0){
		sdlFehlerAusgeben("Warnung: Tonausgabe konnte nicht auf "
				"44100Hz 16-bit Stereo gesetzt werden.");
	}

	if (SDL_Init(SDL_INIT_JOYSTICK) != 0){
		sdlFehlerAusgeben("Warnung: SDL-Joystick konnte nicht initialisiert werden.");
	}

#ifndef OUTPUT_SDL
	/* OpenGL-Bildschirmmodus setzen */

	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT + HUD_HOEHE, 0, SDL_OPENGL);
#else
	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT + HUD_HOEHE, 32, SDL_HWSURFACE | SDL_DOUBLEBUF);
#endif

	if (screen == NULL) {
		sdlFehlerAusgeben("Konnte Bildschirmmodus nicht setzen.");
		exit(1);
	}

#ifndef OUTPUT_SDL
	/* Initialisierung von OpenGL */

	// OpenGL-Anzeigeflche auf das gesamte Fenster ausdehnen.
	glViewport(0, 0, screen->w, screen->h);

	// Einheitsmatrix in die Projektionsmatrix laden und
	// das Koordinatensystem so setzen, dass (0, 0)
	// die linke obere Ecke darstellt.
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, screen->w, screen->h, 0, -1.0, 1.0);

	// Einheitsmatrix in die Modelmatrix laden.
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Alphatest aktivieren, um Bildteile mit einem Alphawert von
	// 0 ganz zu ignorieren, dies ist effizienter.
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0);

	// Alphablending aktivieren.
	glEnable(GL_BLEND);
	// Standardalphablending einstellen, das jeweils untere Objekt wird
	// als Hintergrund betrachtet.
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Volle Helligkeit und keine Einfrbung der Texturen.
	glColor3ub(255, 255, 255);
	
	// Rckseiten von Polygonen auch zeichnen.
	glDisable(GL_CULL_FACE);
#endif

	mainScreen = screen;
	return screen;
}

std::vector<std::string> parse(std::string input_, int anzahl)
{
	std::istringstream input(input_);
	std::vector<std::string> output;
	std::string tmp;

	for (int i = 0; i < anzahl; i++){
		input >> tmp;
		output.push_back(tmp);
	}

	// Rest
	getline(input, tmp);
	output.push_back(tmp);
	
	return output;
}

void zeichneRechteck(int x, int y, int breite, int hoehe)
{
#ifndef OUTPUT_SDL
	GL_TexEnable(0);				
	glBegin(GL_QUADS);
		glVertex2i(x,          y     );
		glVertex2i(x + breite, y     );
		glVertex2i(x + breite, y + hoehe);
		glVertex2i(x,          y + hoehe);
	glEnd();
#else
	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = breite;
	rect.h = hoehe;

	SDL_FillRect(mainScreen, &rect, SDL_MapRGB(mainScreen->format, 255, 255, 255));
#endif
}

void zeichneHohlRechteck(int x, int y, int breite, int hoehe)
{
	zeichneRechteck(x,              y,          breite - 1, 1        );
	zeichneRechteck(x + breite - 1, y,          1,          hoehe - 1);
	zeichneRechteck(x,              y + hoehe - 1,          breite, 1);
	zeichneRechteck(x,              y,          1,          hoehe - 1);
}

void trim(std::string& str)
{
	std::string::size_type start = str.find_first_not_of(' ');
	std::string::size_type end = str.find_last_not_of(' ');
	if (start == std::string::npos){
		start = 0;
	}
	if (end == std::string::npos){
		end = str.length() - 1;
	}
	str = str.substr(start, (end - start) + 1);
}

std::vector<std::string> ladeStringVectorAusDatei(std::string dateiname,
			bool leereZeilenIgnorieren)
{
	std::vector<std::string> output;
	std::ifstream datei(dateiname.c_str());

	if (datei == NULL){
		throw "Datei '" + dateiname + "' konnte nicht geladen werden.";
	}

	std::string zeile;

	while(getline(datei, zeile)){
		trim(zeile);
		if ( !leereZeilenIgnorieren || (zeile != "") ){
			output.push_back(zeile);
		}
	}
	datei.close();

	return output;
}

std::string itos(int wert)
{
	std::stringstream stream;
	std::string output;
	stream << wert;
	stream >> output;
	return output;
}
